from abc import ABC, abstractmethod
import itertools
import logging
import multiprocessing
import os
import signal
import time
from typing import Any, AsyncGenerator, List, NamedTuple, Optional, Tuple  # noqa: F401

from lahja import BroadcastConfig, ConnectionConfig
from lahja.base import EndpointAPI
from lahja.tools.benchmark.backends import BaseBackend
from lahja.tools.benchmark.constants import (
    DRIVER_ENDPOINT,
    REPORTER_ENDPOINT,
    ROOT_ENDPOINT,
)
from lahja.tools.benchmark.logging import setup_stderr_lahja_logging
from lahja.tools.benchmark.stats import GlobalStatistic, LocalStatistic
from lahja.tools.benchmark.typing import (
    PerfMeasureEvent,
    PerfMeasureRequest,
    PerfMeasureResponse,
    RawMeasureEntry,
    ShutdownEvent,
    TotalRecordedEvent,
)
from lahja.tools.benchmark.utils.reporting import print_full_report


class DriverProcessConfig(NamedTuple):
    num_events: int
    connected_endpoints: Tuple[ConnectionConfig, ...]
    throttle: float
    payload_bytes: int
    backend: BaseBackend
    debug_logging: bool


class BaseDriverProcess(ABC):
    logger = logging.getLogger("lahja.tools.benchmark.process.DriverProcess")

    def __init__(self, config: DriverProcessConfig) -> None:
        self._config = config
        self._process: Optional[multiprocessing.Process] = None

    def start(self) -> None:
        self._process = multiprocessing.Process(
            target=self.launch, args=(self._config,), daemon=True
        )
        self._process.start()

    def stop(self) -> None:
        if self._process is None:
            raise Exception("no process")
        elif self._process.pid is not None:
            os.kill(self._process.pid, signal.SIGINT)
        else:
            self._process.terminate()

        try:
            self._process.join(1)
        except TimeoutError:
            self._process.terminate()
            self._process.join(1)

    @classmethod
    def launch(cls, config: DriverProcessConfig) -> None:
        if config.debug_logging:
            setup_stderr_lahja_logging()

        try:
            config.backend.run(cls.worker, config)
        except KeyboardInterrupt:
            return

    @classmethod
    async def worker(cls, config: DriverProcessConfig) -> None:
        conn_config = ConnectionConfig.from_name(DRIVER_ENDPOINT)
        async with config.backend.Endpoint.serve(conn_config) as event_bus:
            await event_bus.connect_to_endpoints(*config.connected_endpoints)

            await cls.do_driver(event_bus, config)

    @staticmethod
    @abstractmethod
    async def do_driver(event_bus: EndpointAPI, config: DriverProcessConfig) -> None:
        ...


class BroadcastDriver(BaseDriverProcess):
    @staticmethod
    async def do_driver(event_bus: EndpointAPI, config: DriverProcessConfig) -> None:
        for consumer in config.connected_endpoints:
            await event_bus.wait_until_endpoint_subscribed_to(
                consumer.name, PerfMeasureEvent
            )

        counter = itertools.count()
        payload = b"\x00" * config.payload_bytes
        while True:
            await config.backend.sleep(config.throttle)
            await event_bus.broadcast(
                PerfMeasureEvent(payload, next(counter), time.time())
            )


class RequestDriver(BaseDriverProcess):
    @classmethod
    async def do_driver(
        cls, event_bus: EndpointAPI, config: DriverProcessConfig
    ) -> None:
        for consumer in config.connected_endpoints:
            await event_bus.wait_until_endpoint_subscribed_to(
                consumer.name, PerfMeasureRequest
            )

        counter = itertools.count()
        payload = b"\x00" * config.payload_bytes
        while True:
            await config.backend.sleep(config.throttle)
            await event_bus.request(
                PerfMeasureRequest(payload, next(counter), time.time())
            )


class ConsumerConfig(NamedTuple):
    num_events: int
    backend: BaseBackend
    debug_logging: bool


class BaseConsumerProcess(ABC):
    logger = logging.getLogger("lahja.tools.benchmark.process.ConsumerProcess")

    def __init__(self, name: str, config: ConsumerConfig) -> None:
        self._name = name
        self._config = config
        self._process: Optional[multiprocessing.Process] = None

    def start(self) -> None:
        self._process = multiprocessing.Process(
            target=self.launch, args=(self._name, self._config)
        )
        self._process.start()

    @classmethod
    def launch(cls, name: str, config: ConsumerConfig) -> None:
        if config.debug_logging:
            setup_stderr_lahja_logging()

        config.backend.run(cls.worker, name, config)

    @classmethod
    async def worker(cls, name: str, config: ConsumerConfig) -> None:
        conn_config = ConnectionConfig.from_name(name)
        async with config.backend.Endpoint.serve(conn_config) as event_bus:
            await event_bus.connect_to_endpoints(
                ConnectionConfig.from_name(REPORTER_ENDPOINT)
            )
            await event_bus.wait_until_connected_to(DRIVER_ENDPOINT)
            stats = await cls.do_consumer(event_bus, config)

            await event_bus.wait_until_endpoint_subscribed_to(
                REPORTER_ENDPOINT, TotalRecordedEvent
            )

            await event_bus.broadcast(
                TotalRecordedEvent(stats.crunch(event_bus.name)),
                BroadcastConfig(filter_endpoint=REPORTER_ENDPOINT),
            )

    @staticmethod
    @abstractmethod
    async def do_consumer(
        event_bus: EndpointAPI, config: ConsumerConfig
    ) -> LocalStatistic:
        ...


class BroadcastConsumer(BaseConsumerProcess):
    @staticmethod
    async def do_consumer(
        event_bus: EndpointAPI, config: ConsumerConfig
    ) -> LocalStatistic:
        stats = LocalStatistic()
        events = event_bus.stream(PerfMeasureEvent, num_events=config.num_events)
        async for event in events:
            stats.add(RawMeasureEntry(sent_at=event.sent_at, received_at=time.time()))
        return stats


class RequestConsumer(BaseConsumerProcess):
    @staticmethod
    async def do_consumer(
        event_bus: EndpointAPI, config: ConsumerConfig
    ) -> LocalStatistic:
        stats = LocalStatistic()
        events = event_bus.stream(PerfMeasureRequest, num_events=config.num_events)
        async for event in events:
            await event_bus.broadcast(PerfMeasureResponse(), event.broadcast_config())
            stats.add(RawMeasureEntry(sent_at=event.sent_at, received_at=time.time()))
        return stats


class ReportingProcessConfig(NamedTuple):
    num_processes: int
    num_events: int
    throttle: float
    payload_bytes: int
    backend: BaseBackend
    debug_logging: bool


class ReportingProcess:
    logger = logging.getLogger("lahja.tools.benchmark.process.ReportingProcess")

    def __init__(self, config: ReportingProcessConfig) -> None:
        self._name = REPORTER_ENDPOINT
        self._config = config
        self._process: Optional[multiprocessing.Process] = None

    def start(self) -> None:
        self._process = multiprocessing.Process(
            target=self.launch, args=(self._config,)
        )
        self._process.start()

    @classmethod
    def launch(cls, config: ReportingProcessConfig) -> None:
        if config.debug_logging:
            setup_stderr_lahja_logging()

        logging.basicConfig(level=logging.INFO, format="%(message)s")
        logger = logging.getLogger("reporting")

        config.backend.run(ReportingProcess.worker, logger, config)

    @staticmethod
    async def worker(logger: logging.Logger, config: ReportingProcessConfig) -> None:
        conn_config = ConnectionConfig.from_name(REPORTER_ENDPOINT)
        async with config.backend.Endpoint.serve(conn_config) as event_bus:
            await event_bus.connect_to_endpoints(
                ConnectionConfig.from_name(ROOT_ENDPOINT)
            )

            global_statistic = GlobalStatistic()
            events = event_bus.stream(
                TotalRecordedEvent, num_events=config.num_processes
            )
            async for event in events:
                global_statistic.add(event.total)

            print_full_report(
                logger,
                config.backend,
                config.num_processes,
                config.num_events,
                global_statistic,
            )
            await event_bus.broadcast(
                ShutdownEvent(), BroadcastConfig(filter_endpoint=ROOT_ENDPOINT)
            )



import asyncio
import collections
import itertools
import logging
import pathlib
import pickle
from typing import (
    Any,
    AsyncGenerator,
    AsyncIterator,
    Awaitable,
    Callable,
    DefaultDict,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from async_generator import asynccontextmanager
import trio
import trio_typing

from lahja import constants
from lahja.base import (
    BaseEndpoint,
    BaseRemoteEndpoint,
    ConnectionAPI,
    EndpointAPI,
    RemoteEndpointAPI,
    TResponse,
    TStreamEvent,
    TSubscribeEvent,
)
from lahja.common import (
    BaseEvent,
    BaseRequestResponseEvent,
    Broadcast,
    BroadcastConfig,
    ConnectionConfig,
    Message,
    Msg,
    Subscription,
    should_endpoint_receive_item,
)
from lahja.exceptions import (
    ConnectionAttemptRejected,
    LifecycleError,
    RemoteDisconnected,
    UnexpectedResponse,
)
from lahja.typing import ConditionAPI, RequestID

from .future import Future


class TrioConnection(ConnectionAPI):
    logger = logging.getLogger("lahja.trio.Connection")

    def __init__(self, socket: trio.SocketStream) -> None:
        self._socket = socket
        self._msg_send_channel, self._msg_receive_channel = cast(
            Tuple[trio.abc.SendChannel[Message], trio.abc.ReceiveChannel[Message]],
            trio.open_memory_channel(100),
        )
        self._write_lock = trio.Lock()
        super().__init__()

    def __str__(self) -> str:
        return f"TrioConnection[{self._socket}]"

    def __repr__(self) -> str:
        return f"<{self}>"

    async def close(self) -> None:
        await self._socket.aclose()

    #
    # Connection API
    #
    @classmethod
    async def connect_to(cls, path: pathlib.Path) -> "TrioConnection":
        socket = await trio.open_unix_socket(str(path))
        cls.logger.debug("Opened connection to %s: %s", path, socket)
        return cls(socket)

    async def send_message(self, message: Msg) -> None:
        msg_data = pickle.dumps(message, protocol=pickle.HIGHEST_PROTOCOL)
        size = len(msg_data)
        try:
            async with self._write_lock:
                await self._socket.send_all(size.to_bytes(4, "little") + msg_data)
        except (trio.ClosedResourceError, trio.BrokenResourceError) as err:
            raise RemoteDisconnected from err

    async def read_message(self) -> Message:
        buffer = bytearray()

        while len(buffer) < 4:
            try:
                data = await self._socket.receive_some(4 - len(buffer))
            except (trio.ClosedResourceError, trio.BrokenResourceError) as err:
                raise RemoteDisconnected from err

            if data == b"":
                raise RemoteDisconnected()

            buffer.extend(data)

        t_size = 4 + int.from_bytes(buffer[:4], "little")

        while len(buffer) < t_size:
            try:
                data = await self._socket.receive_some(t_size - len(buffer))
            except (trio.ClosedResourceError, trio.BrokenResourceError) as err:
                raise RemoteDisconnected from err

            if data == b"":
                raise RemoteDisconnected()

            buffer.extend(data)

        msg = cast(Message, pickle.loads(buffer[4:t_size]))
        return msg


class TrioRemoteEndpoint(BaseRemoteEndpoint):
    def __init__(
        self,
        local_name: str,
        conn: ConnectionAPI,
        subscriptions_changed: ConditionAPI,
        new_msg_func: Callable[[Broadcast], Awaitable[Any]],
    ) -> None:
        super().__init__(local_name, conn, new_msg_func)

        self._notify_lock = trio.Lock()  # type: ignore

        self._received_response = trio.Condition()  # type: ignore
        self._received_subscription = trio.Condition()  # type: ignore

        self._running = trio.Event()  # type: ignore
        self._stopped = trio.Event()  # type: ignore

        self._received_subscription = subscriptions_changed

        self._subscriptions_initialized = trio.Event()  # type: ignore

        self._running = trio.Event()  # type: ignore
        self._stopped = trio.Event()  # type: ignore
        self._ready = trio.Event()  # type: ignore

    @asynccontextmanager
    async def run(self) -> AsyncIterator[RemoteEndpointAPI]:
        async with trio.open_nursery() as nursery:
            await self._start(nursery)
            try:
                yield self
            finally:
                await self.stop()

    async def _start(self, nursery: trio_typing.Nursery) -> None:
        if self.is_running:
            raise LifecycleError("RemoteEndpoint is already running")
        elif self.is_stopped:
            raise LifecycleError("RemoteEndpoint has already been run and stopped")

        nursery.start_soon(self._process_incoming_messages)
        # this is a trade-off between allowing flexibility in how long it takes
        # for a `Runnable` to start and getting good errors in the event that
        # the `_run` method forgets to set the `_running` event.
        with trio.fail_after(2):
            await self.wait_started()

    async def stop(self) -> None:
        if self.is_stopped:
            return
        self._stopped.set()
        await self.conn.close()


async def _wait_for_path(path: trio.Path) -> None:
    """
    Wait for the path to appear at ``path``
    """
    while not await path.exists():
        await trio.sleep(0.05)


WireBroadcastChannelPair = Tuple[
    trio.abc.SendChannel[Broadcast], trio.abc.ReceiveChannel[Broadcast]
]

OutboundBroadcast = Tuple[
    Optional[trio.Event], BaseEvent, Optional[BroadcastConfig], Optional[RequestID]
]
OutboundBroadcastChannelPair = Tuple[
    trio.abc.SendChannel[OutboundBroadcast], trio.abc.ReceiveChannel[OutboundBroadcast]
]

ConnectionChannelPair = Tuple[
    trio.abc.SendChannel[ConnectionConfig], trio.abc.ReceiveChannel[ConnectionConfig]
]
StreamChannelPair = Tuple[
    trio.abc.SendChannel[BaseEvent], trio.abc.ReceiveChannel[BaseEvent]
]


class TrioEndpoint(BaseEndpoint):
    _stream_channels: DefaultDict[Type[BaseEvent], Set[trio.abc.SendChannel[BaseEvent]]]
    _pending_requests: Dict[RequestID, Future[BaseEvent]]

    # Number of unaccepted connections that we allow before refusing new ones.
    _connection_backlog = 10

    _sync_handlers: DefaultDict[Type[BaseEvent], List[Callable[[TSubscribeEvent], Any]]]
    _async_handlers: DefaultDict[
        Type[BaseEvent], List[Callable[[TSubscribeEvent], Awaitable[Any]]]
    ]

    logger = logging.getLogger("lahja.trio.TrioEndpoint")

    _subscriptions_changed: trio.Event

    def __init__(self, name: str):
        super().__init__(name)

        self._running = trio.Event()
        self._stopped = trio.Event()

        # Temporary storage for SendChannels used in the `stream` API.  Uses a
        # `WeakSet` to automate cleanup.
        self._stream_channels = collections.defaultdict(set)
        self._pending_requests = {}
        self._sync_handlers = collections.defaultdict(list)
        self._async_handlers = collections.defaultdict(list)

        self._run_lock = trio.Lock()

        # Signal when a new remote connection is established
        self._remote_connections_changed = trio.Condition()  # type: ignore

        # Signal when at least one remote has had a subscription change.
        self._remote_subscriptions_changed = trio.Condition()  # type: ignore

        # events used to signal that the endpoint has fully booted up.
        self._message_processing_loop_running = trio.Event()
        self._connection_loop_running = trio.Event()
        self._process_broadcasts_running = trio.Event()

        # internal signal that local subscriptions have changed.
        self._subscriptions_changed = trio.Event()

        self._socket_bound = trio.Event()
        self._server_stopped = trio.Event()

    #
    # Running and endpoint
    #
    @property
    def is_running(self) -> bool:
        return not self.is_stopped and self._running.is_set()

    @property
    def is_stopped(self) -> bool:
        return self._stopped.is_set()

    async def wait_started(self) -> None:
        await self._running.wait()

    async def wait_stopped(self) -> None:
        await self._stopped.wait()

    @asynccontextmanager
    async def run(self) -> AsyncGenerator[EndpointAPI, None]:
        async with trio.open_nursery() as nursery:
            await self._start(nursery)
            try:
                yield self
            finally:
                await self._stop()

    async def _start(self, nursery: trio_typing.Nursery) -> None:
        if self.is_running:
            raise LifecycleError("Endpoint is already running")
        elif self.is_stopped:
            raise LifecycleError("Endpoint has already been run and stopped")

        with trio.fail_after(2):
            await nursery.start(self._run)

            # this is a trade-off between allowing flexibility in how long it takes
            # for a `Runnable` to start and getting good errors in the event that
            # the `_run` method forgets to set the `_running` event.
            await self.wait_started()

    async def _stop(self) -> None:
        if self.is_stopped:
            return
        self._stopped.set()
        await self._cleanup()

    async def _run(self, task_status: trio_typing.TaskStatus[None]) -> None:
        # This send/receive channel is fed any event/config pair that should be
        # broadcast.
        # Those messages are then retrieved from the receive channel by the
        # `_process_outbound_messages` daemon and broadcast to the appropriate
        # `RemoteEndpoint`.
        (self._outbound_send_channel, outbound_receive_channel) = cast(
            OutboundBroadcastChannelPair, trio.open_memory_channel(100)
        )

        # This send/receive channel is fed by connected `RemoteEndpoint`
        # objects which feed received events into the send side of the channel.
        # Those messages are then retrieved from the receive channel by the
        # `_process_inbound_messages` daemon.
        (self._inbound_send_channel, inbound_receive_channel) = cast(
            WireBroadcastChannelPair, trio.open_memory_channel(100)
        )

        # This send/receive channel is fed by
        # `Endpoint.connect_to_endpoint` which places a 2-tuple of
        # (ConnectionConfig, trio.Event) which is retrieved by
        # `_process_connections`.
        (self._connection_send_channel, connection_receive_channel) = cast(
            ConnectionChannelPair, trio.open_memory_channel(100)
        )

        self.logger.debug("%s: starting", self)

        async with trio.open_nursery() as nursery:
            #
            # _process_outbound_messages:
            #     Manages a channel which all outgoing
            #     event broadcasts are placed on, running them each through the
            #     appropriate `RemoteEndpoint.send_message`
            #
            nursery.start_soon(
                self._process_outbound_messages, outbound_receive_channel
            )

            #
            # _process_inbound_messages:
            #     Manages a channel which all incoming
            #     event broadcasts are placed on, running each event through
            #     the internal `_process_item` handler which handles all of the
            #     internal logic for the various
            #     `request/response/subscribe/stream/wait_for` API.
            #
            nursery.start_soon(self._process_inbound_messages, inbound_receive_channel)

            #
            # _process_connections:
            #     When `Endpoint.connect_to_endpoint` is called, the actual
            #     connection process is done asynchronously by putting the
            #     `ConnectionConfig` onto a queue which this process
            #     retrieves to establish the new connection.  This includes
            #     handing the `RemoteEndpoint` oblect off to the `RemoteManager`
            #     which takes care of the connection lifecycle.
            #
            nursery.start_soon(
                self._process_connections, connection_receive_channel, nursery
            )

            #
            # _monitor_subscription_changes
            #    Monitors an event for local changes to subscriptions to
            #    propagate those changes to remotes.
            nursery.start_soon(self._monitor_subscription_changes)

            # mark the endpoint as running.
            self._running.set()
            # tell the nursery that we are started.
            task_status.started()

            await self.wait_stopped()

            nursery.cancel_scope.cancel()

    async def _cleanup(self) -> None:
        # Cleanup stateful things.
        await self._outbound_send_channel.aclose()
        await self._inbound_send_channel.aclose()
        await self._connection_send_channel.aclose()

        del self._outbound_send_channel
        del self._inbound_send_channel
        del self._connection_send_channel

        self.logger.debug("%s: stopped", self)

    #
    # Background processes
    #
    async def _process_outbound_messages(
        self, channel: trio.abc.ReceiveChannel[OutboundBroadcast]
    ) -> None:
        """
        Consume events that have been received from a remote endpoint and
        process them.
        """
        async for (done, item, config, id) in channel:
            item.bind(self, id)

            is_eligible = should_endpoint_receive_item(
                item, config, self.name, self.get_subscribed_events()
            )
            is_internal = config is not None and config.internal

            if is_eligible:
                await self._process_item(item, config)

            if is_internal:
                # if the event is flagged as internal we exit early since it should
                # not be broadcast beyond this endpoint
                return

            remotes_for_broadcast = tuple(
                remote
                for remote in self._connections
                if should_endpoint_receive_item(
                    item, config, remote.name, remote.get_subscribed_events()
                )
            )

            compressed_item = self._compress_event(item)
            message = Broadcast(compressed_item, config)
            for remote in remotes_for_broadcast:
                try:
                    await remote.send_message(message)
                except RemoteDisconnected as err:
                    self.logger.debug(
                        "%s: dropping disconnected remote %s: %s", remote, err
                    )
                    await remote.stop()

            # This Future is used to signal back to the `broadcast` method that
            # this event has been broadcast and it is safe to return from the
            # method.
            if done is not None:
                done.set()

    async def _process_inbound_messages(
        self, channel: trio.abc.ReceiveChannel[Broadcast]
    ) -> None:
        """
        Consume events that have been received from a remote endpoint and
        process them.
        """
        async for (item, config) in channel:
            event = self._decompress_event(item)
            await self._process_item(event, config)

    async def _process_connections(
        self,
        channel: trio.abc.ReceiveChannel[ConnectionConfig],
        nursery: trio_typing.Nursery,
    ) -> None:
        """
        Long running process that establishes connections to endpoint servers
        and runs the handler for receiving events sent by the server over that
        connection.
        """
        self.logger.debug("%s: starting new connection channel", self)
        self._connection_loop_running.set()
        async for config in channel:
            # Allow some time for for the IPC socket to appear
            with trio.fail_after(constants.IPC_WAIT_SECONDS):
                await _wait_for_path(trio.Path(config.path))

            await trio.sleep(0.001)
            # Establish the socket connection
            connection = await TrioConnection.connect_to(config.path)

            # Create the remote
            remote = TrioRemoteEndpoint(
                self.name,
                connection,
                self._remote_subscriptions_changed,
                self._inbound_send_channel.send,
            )
            nursery.start_soon(self._run_remote_endpoint, remote)

    async def _monitor_subscription_changes(self) -> None:
        while not self.is_stopped:
            # We wait for the event to change and then immediately replace it
            # with a new event.  This **must** occur before any additional
            # `await` calls to ensure that any *new* changes to the
            # subscriptions end up operating on the *new* event and will be
            # picked up in the next iteration of the loop.
            await self._subscriptions_changed.wait()
            self._subscriptions_changed = trio.Event()

            # make a copy so that the set doesn't change while we iterate
            # over it
            subscribed_events = self.get_subscribed_events()

            async with trio.open_nursery() as nursery:
                async with self._remote_connections_changed:
                    for remote in self._connections:
                        nursery.start_soon(
                            remote.notify_subscriptions_updated,
                            subscribed_events,
                            False,
                        )
            async with self._remote_subscriptions_changed:
                self._remote_subscriptions_changed.notify_all()

    async def _process_item(
        self, item: BaseEvent, config: Optional[BroadcastConfig]
    ) -> None:
        event_type = type(item)

        # handle request/response
        if config is not None and config.filter_event_id in self._pending_requests:
            fut = self._pending_requests.pop(config.filter_event_id)
            fut.set_result(item)

        # handle stream channel
        if event_type in self._stream_channels:
            channels = tuple(self._stream_channels[event_type])
            for send_channel in channels:
                try:
                    await send_channel.send(item)
                except trio.ClosedResourceError:
                    self._stream_channels[event_type].remove(send_channel)

        # handle subscriptions
        if event_type in self._sync_handlers:
            for handler_fn in self._sync_handlers[event_type]:
                try:
                    handler_fn(item)
                except Exception as err:
                    self.logger.debug(
                        "%s: handler function %s error: %s", self, handler_fn, err
                    )
        if event_type in self._async_handlers:
            for handler_fn in self._async_handlers[event_type]:
                try:
                    await handler_fn(item)
                except Exception as err:
                    self.logger.debug(
                        "%s: handler function %s error: %s", self, handler_fn, err
                    )

    #
    # Server API
    #
    @property
    def is_serving(self) -> bool:
        # NOTE: This does not guarantee that we are actually serving requests since _run_server()
        # does not have a way of notifying us when trio.serve_listeners() has actually been
        # scheduled.
        return not self.is_server_stopped and self._socket_bound.is_set()

    @property
    def is_server_stopped(self) -> bool:
        return self._server_stopped.is_set()

    @classmethod
    @asynccontextmanager
    async def serve(cls, config: ConnectionConfig) -> AsyncIterator["TrioEndpoint"]:
        endpoint = cls(config.name)
        async with endpoint.run():
            async with trio.open_nursery() as nursery:
                await endpoint._start_serving(nursery, config.path)
                try:
                    yield endpoint
                finally:
                    await endpoint._stop_serving()

    async def _start_serving(
        self, nursery: trio_typing.Nursery, ipc_path: pathlib.Path
    ) -> None:
        if not self.is_running:
            raise LifecycleError("Cannot start server if endpoint is not running")
        elif self.is_stopped:
            raise LifecycleError("Endpoint has already been run and stopped")
        elif self.is_serving:
            raise LifecycleError("Endpoint is already serving")
        elif self.is_server_stopped:
            raise LifecycleError("Endpoint server already ran and was stopped")

        self.ipc_path = ipc_path
        nursery.start_soon(self._run_server)

        # Wait until the ipc socket has appeared and is accepting connections.
        with trio.fail_after(constants.IPC_WAIT_SECONDS):
            await _wait_for_path(trio.Path(ipc_path))

        await self._socket_bound.wait()

    async def _stop_serving(self) -> None:
        if self.is_server_stopped:
            return

        self._server_stopped.set()
        self._server_nursery.cancel_scope.cancel()

        del self._server_nursery

        try:
            self.ipc_path.unlink()
        except FileNotFoundError:
            pass
        self.logger.debug(f"%s: server stopped", self)

    async def _run_server(self) -> None:
        async with trio.open_nursery() as nursery:
            # Store nursery on self so that we can access it for cancellation
            self._server_nursery = nursery

            self.logger.debug("%s: server starting", self)
            socket = trio.socket.socket(trio.socket.AF_UNIX, trio.socket.SOCK_STREAM)
            await socket.bind(self.ipc_path.__fspath__())
            socket.listen(self._connection_backlog)
            listener = trio.SocketListener(socket)

            async def set_socket_bound() -> None:
                self._socket_bound.set()

            # Use start_soon here so that we give serve_listeners() below a chance to run before
            # other endpoints start connecting to us.
            nursery.start_soon(set_socket_bound)

            try:
                await trio.serve_listeners(
                    handler=self._accept_conn,
                    listeners=(listener,),
                    handler_nursery=nursery,
                )
            finally:
                self.logger.debug("%s: server finished", self)

    async def _accept_conn(self, socket: trio.SocketStream) -> None:
        self.logger.debug("%s: starting client handler for %s", self, socket)
        connection = TrioConnection(socket)
        remote = TrioRemoteEndpoint(
            self.name,
            connection,
            self._remote_subscriptions_changed,
            self._inbound_send_channel.send,
        )
        await self._run_remote_endpoint(remote)

    #
    # Establishing connections
    #
    async def connect_to_endpoints(self, *endpoints: ConnectionConfig) -> None:
        """
        Connect to the given endpoints and await until all connections are established.
        """
        if not self.is_running:
            raise ConnectionAttemptRejected(
                "Cannot establish connections if endpoint isn't running"
            )

        for config in endpoints:
            # Ensure we are not already connected to the named endpoint
            if self.is_connected_to(config.name):
                raise ConnectionAttemptRejected(
                    f"Already connected to endpoint with name: {config.name}"
                )

            # Feed the `ConnectionConfig` through a channel where
            # `_process_connections` will pick it up and actually establish the
            # connection.
            await self._connection_send_channel.send(config)

        for config in endpoints:
            with trio.fail_after(constants.ENDPOINT_CONNECT_TIMEOUT):
                await self.wait_until_connected_to(config.name)

    #
    # Primary endpoint API
    #
    async def broadcast(
        self, item: BaseEvent, config: Optional[BroadcastConfig] = None
    ) -> None:
        """
        Broadcast an instance of :class:`~lahja.common.BaseEvent` on the event bus. Takes
        an optional second parameter of :class:`~lahja.common.BroadcastConfig` to decide
        where this event should be broadcasted to. By default, events are broadcasted across
        all connected endpoints with their consuming call sites.
        """
        self.maybe_raise_no_subscribers_exception(config, type(item))
        done = trio.Event()
        await self._outbound_send_channel.send((done, item, config, None))
        await done.wait()

    def broadcast_nowait(
        self, item: BaseEvent, config: Optional[BroadcastConfig] = None
    ) -> None:
        self.maybe_raise_no_subscribers_exception(config, type(item))
        # FIXME: Ignoring type check because of https://github.com/python-trio/trio/issues/1327
        self._outbound_send_channel.send_nowait(  # type: ignore
            (None, item, config, None)
        )

    TResponse = TypeVar("TResponse", bound=BaseEvent)

    async def request(
        self,
        item: BaseRequestResponseEvent[TResponse],
        config: Optional[BroadcastConfig] = None,
    ) -> TResponse:
        """
        Broadcast an instance of
        :class:`~lahja.common.BaseRequestResponseEvent` on the event bus and
        immediately wait on an expected answer of type
        :class:`~lahja.common.BaseEvent`. Optionally pass a second parameter of
        :class:`~lahja.common.BroadcastConfig` to decide where the request
        should be broadcasted to. By default, requests are broadcasted across
        all connected endpoints with their consuming call sites.
        """
        self.maybe_raise_no_subscribers_exception(config, type(item))
        request_id = next(self._get_request_id)

        # Create an asynchronous generator that we use to pipe the result
        fut: Future[TResponse] = Future()  # type: ignore

        # place the send channel where the message processing loop can find it.
        self._pending_requests[request_id] = fut

        await self._outbound_send_channel.send((None, item, config, request_id))

        # await for the result to be sent through the channel.
        result: TResponse = await fut  # type: ignore
        expected_response_type = item.expected_response_type()
        if not isinstance(result, expected_response_type):
            raise UnexpectedResponse(
                f"The type of the response is {type(result)}, expected: "
                f"{expected_response_type}"
            )

        return result

    TSubscribeEvent = TypeVar("TSubscribeEvent", bound=BaseEvent)

    def subscribe(
        self,
        event_type: Type[TSubscribeEvent],
        handler: Callable[[TSubscribeEvent], Union[Any, Awaitable[Any]]],
    ) -> Subscription:
        """
        Subscribe to receive updates for any event that matches the specified event type.
        A handler is passed as a second argument an :class:`~lahja.common.Subscription` is returned
        to unsubscribe from the event if needed.
        """
        if asyncio.iscoroutinefunction(handler):
            self._async_handlers[event_type].append(handler)
            subscription = Subscription(
                lambda: self._async_handlers[event_type].remove(handler)
            )
        else:
            self._sync_handlers[event_type].append(handler)
            subscription = Subscription(
                lambda: self._sync_handlers[event_type].remove(handler)
            )

        # notify subscriptions have been updated.
        self._subscriptions_changed.set()

        return subscription

    TStreamEvent = TypeVar("TStreamEvent", bound=BaseEvent)

    async def stream(
        self, event_type: Type[TStreamEvent], num_events: Optional[int] = None
    ) -> AsyncGenerator[TStreamEvent, None]:
        """
        Stream all events that match the specified event type. This returns an
        ``AsyncIterable[BaseEvent]`` which can be consumed through an ``async for`` loop.
        An optional ``num_events`` parameter can be passed to stop streaming after a maximum amount
        of events was received.
        """
        (send_channel, receive_channel) = cast(
            StreamChannelPair, trio.open_memory_channel(100)
        )

        self._stream_channels[event_type].add(send_channel)

        # notify subscriptions have been updated.
        self._subscriptions_changed.set()

        if num_events is None:
            # iterate forever
            counter = itertools.count()
        else:
            # fixed number of iterations
            counter = iter(range(max(0, num_events - 1)))

        async for event in receive_channel:
            yield event  # type: ignore  # mypy doesn't recognize this having a correct type

            try:
                next(counter)
            except StopIteration:
                await send_channel.aclose()
                break

        # We only need to trigger a subscription change if there are no more
        # streams for this event.
        self._stream_channels[event_type].remove(send_channel)
        if not self._stream_channels[event_type]:
            self._stream_channels.pop(event_type)
            self._subscriptions_changed.set()

    #
    # Subscriptions API
    #
    def get_subscribed_events(self) -> Set[Type[BaseEvent]]:
        """
        Return the set of events this Endpoint is currently listening for
        """
        return (
            set(self._sync_handlers.keys())
            .union(self._async_handlers.keys())
            .union(self._stream_channels.keys())
        )



import functools
import logging
from typing import Any, AsyncContextManager, Callable

from lahja import ConnectionConfig, EndpointAPI
from lahja.tools.engine import EngineAPI

logger = logging.getLogger("lahja.tools.initializers")


class Initializer:
    def __str__(self) -> str:
        return self._fn.__name__

    def __init__(
        self, fn: Callable[..., AsyncContextManager[EndpointAPI]], **kwargs: Any
    ) -> None:
        functools.update_wrapper(self, fn)
        self._fn = fn
        self._kwargs = kwargs

    def __call__(self, engine: EngineAPI) -> AsyncContextManager[EndpointAPI]:
        return self._fn(engine, **self._kwargs)


#
# EndpointAPI.serve
#
def _serve_endpoint(
    engine: EngineAPI, config: ConnectionConfig
) -> AsyncContextManager[EndpointAPI]:
    logger.debug(
        "[%s(%s)].serve(%s)", engine.endpoint_class.__name__, config.name, config.path
    )
    return engine.endpoint_class.serve(config)


def serve_endpoint(config: ConnectionConfig) -> Initializer:
    return Initializer(_serve_endpoint, config=config)


#
# EndpointAPI.run
#
def _run_endpoint(engine: EngineAPI, name: str) -> AsyncContextManager[EndpointAPI]:
    logger.debug("[%s(%s)].run()", engine.endpoint_class.__name__, name)
    # EndpointAPI doesn't specify an __init__ so mypy doesn't understand this.
    return engine.endpoint_class(name).run()  # type: ignore


def run_endpoint(name: str) -> Initializer:
    return Initializer(_run_endpoint, name=name)



from abc import ABC, abstractmethod
import itertools
from pathlib import Path
from typing import (  # noqa: F401
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Iterator,
    NamedTuple,
    Optional,
    Set,
    Type,
    TypeVar,
    Union,
)

from lahja.exceptions import BindError
from lahja.typing import RequestID

if TYPE_CHECKING:
    from lahja.base import EndpointAPI  # noqa: F401


class Subscription:
    def __init__(self, unsubscribe_fn: Callable[[], Any]) -> None:
        self._unsubscribe_fn = unsubscribe_fn

    def unsubscribe(self) -> None:
        self._unsubscribe_fn()


class BroadcastConfig:
    def __init__(
        self,
        filter_endpoint: Optional[str] = None,
        filter_event_id: Optional[RequestID] = None,
        internal: bool = False,
        require_subscriber: bool = True,
    ) -> None:

        self.filter_endpoint = filter_endpoint
        self.filter_event_id = filter_event_id
        self.internal = internal
        self.require_subscriber = require_subscriber

        if self.internal and self.filter_endpoint is not None:
            raise ValueError("`internal` can not be used with `filter_endpoint")

    def __str__(self) -> str:
        return (
            "BroadcastConfig["
            f"{'internal' if self.internal else 'external'} / "
            f"endpoint={self.filter_endpoint if self.filter_endpoint else 'N/A'} / "
            f"  id={self.filter_event_id if self.filter_event_id else 'N/A'}"
            "]"
        )

    def allowed_to_receive(self, endpoint: str) -> bool:
        return self.filter_endpoint is None or self.filter_endpoint == endpoint


class BaseEvent:

    _origin = ""
    _id: Optional[RequestID] = None

    is_bound = False

    def get_origin(self) -> str:
        if not self.is_bound:
            raise AttributeError("Event is not bound")
        return self._origin

    def bind(self, endpoint: "EndpointAPI", id: Optional[RequestID]) -> None:
        if self.is_bound:
            raise BindError("Event is already bound")
        self._origin = endpoint.name
        self._id = id
        self.is_bound = True

    def broadcast_config(self, internal: bool = False) -> BroadcastConfig:
        if internal:
            return BroadcastConfig(internal=True, filter_event_id=self._id)

        return BroadcastConfig(filter_endpoint=self._origin, filter_event_id=self._id)


TResponse = TypeVar("TResponse", bound=BaseEvent)


class BaseRequestResponseEvent(ABC, BaseEvent, Generic[TResponse]):
    @staticmethod
    @abstractmethod
    def expected_response_type() -> Type[TResponse]:
        """
        Return the type that is expected to be send back for this request.
        This ensures that at runtime, only expected responses can be send
        back to callsites that issued a `BaseRequestResponseEvent`
        """
        raise NotImplementedError("Must be implemented by subsclasses")


class ConnectionConfig(NamedTuple):
    """
    Configuration class needed to establish :class:`~lahja.endpoint.Endpoint` connections.
    """

    name: str
    path: Path

    @classmethod
    def from_name(
        cls, name: str, base_path: Optional[Path] = None
    ) -> "ConnectionConfig":
        if base_path is None:
            return cls(name=name, path=Path(f"{name}.ipc"))
        elif base_path.is_dir():
            return cls(name=name, path=base_path / f"{name}.ipc")
        else:
            raise TypeError("Provided `base_path` must be a directory")


class Broadcast(NamedTuple):
    event: Union[BaseEvent, bytes]
    config: Optional[BroadcastConfig]


class Message(ABC):
    """
    Base class for all valid message types that an ``Endpoint`` can handle.
    ``NamedTuple`` breaks multiple inheritance which means, instead of regular subclassing,
    derived message types need to derive from ``NamedTuple`` directly and call
    Message.register(DerivedType) in order to allow isinstance(obj, Message) checks.
    """

    pass


class SubscriptionsUpdated(NamedTuple):
    subscriptions: Set[Type[BaseEvent]]
    response_expected: bool


class SubscriptionsAck:
    pass


class Hello(NamedTuple):
    name: str


Message.register(Broadcast)
Message.register(SubscriptionsUpdated)
Message.register(SubscriptionsAck)
Message.register(Hello)


# mypy doesn't appreciate the ABCMeta trick
Msg = Union[Broadcast, SubscriptionsUpdated, SubscriptionsAck, Hello]


class RequestIDGenerator(Iterator[RequestID]):
    def __init__(self, base: bytes):
        self._base = base
        self._counter = itertools.count()

    def __next__(self) -> RequestID:
        return self._base + next(self._counter).to_bytes(4, "little")  # type: ignore


def should_endpoint_receive_item(
    item: BaseEvent,
    config: Optional[BroadcastConfig],
    endpoint_name: str,
    subscribed_events: Set[Type[BaseEvent]],
) -> bool:
    if config is not None:
        if not config.allowed_to_receive(endpoint_name):
            return False
        elif config.filter_event_id is not None:
            # the item is a response to a request.
            return True

    return type(item) in subscribed_events



class LahjaError(Exception):
    """
    Base class for all lahja errors
    """

    pass


class BindError(LahjaError):
    """
    Raise when an attempt was made to bind an event that is already bound.
    """


class ConnectionAttemptRejected(LahjaError):
    """
    Raised when an attempt was made to connect to an endpoint that is already connected.
    """

    pass


class LifecycleError(LahjaError):
    """
    Raised when attempting to violate the lifecycle of an endpoint such as
    starting an already started endpoint or starting an endpoint that has
    already stopped.
    """

    pass


class NoSubscribers(LahjaError):
    """
    Raised when attempting to send an event or make a request while there are no listeners for the
    specific type of event or request.
    This is a safety check, set ``require_subscriber`` of :class:`~lahja.base.BroadcastConfig`
    to ``False`` to allow propagation without listeners.
    """

    pass


class UnexpectedResponse(LahjaError):
    """
    Raised when the type of a response did not match the ``expected_response_type``.
    """

    pass


class RemoteDisconnected(LahjaError):
    """
    Raise when a remote disconnects while we attempting to read a message.
    """

    pass



import asyncio
from asyncio import StreamReader, StreamWriter
from collections import defaultdict
import functools
import itertools
import logging
from pathlib import Path
import pickle
import time
import traceback
from typing import (  # noqa: F401
    Any,
    AsyncGenerator,
    AsyncIterable,
    AsyncIterator,
    Awaitable,
    Callable,
    DefaultDict,
    Dict,
    Iterable,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from async_generator import asynccontextmanager

from lahja import constants
from lahja.base import (
    BaseEndpoint,
    BaseRemoteEndpoint,
    ConnectionAPI,
    EndpointAPI,
    RemoteEndpointAPI,
    TResponse,
    TStreamEvent,
    TSubscribeEvent,
)
from lahja.common import (
    BaseEvent,
    BaseRequestResponseEvent,
    Broadcast,
    BroadcastConfig,
    ConnectionConfig,
    Message,
    Msg,
    Subscription,
    should_endpoint_receive_item,
)
from lahja.exceptions import (
    ConnectionAttemptRejected,
    RemoteDisconnected,
    UnexpectedResponse,
)
from lahja.typing import ConditionAPI, RequestID


async def wait_for_path(path: Path, timeout: int = constants.IPC_WAIT_SECONDS) -> None:
    """
    Wait for the path to appear at ``path``
    """
    start_at = time.monotonic()
    while time.monotonic() - start_at < timeout:
        if path.exists():
            return
        await asyncio.sleep(0.05)

    raise TimeoutError(f"IPC socket file {path} has not appeared in {timeout} seconds")


class AsyncioConnection(ConnectionAPI):
    logger = logging.getLogger("lahja.endpoint.asyncio.Connection")

    def __init__(self, reader: StreamReader, writer: StreamWriter) -> None:
        self.writer = writer
        self.reader = reader
        self._drain_lock = asyncio.Lock()

    @classmethod
    async def connect_to(cls, path: Path) -> ConnectionAPI:
        reader, writer = await asyncio.open_unix_connection(str(path))
        return cls(reader, writer)

    async def close(self) -> None:
        if not self.reader.at_eof():
            self.reader.feed_eof()
        self.writer.close()

    async def send_message(self, message: Msg) -> None:
        pickled = pickle.dumps(message, protocol=pickle.HIGHEST_PROTOCOL)
        size = len(pickled)

        try:
            self.writer.write(
                size.to_bytes(constants.SIZE_MARKER_LENGTH, "little") + pickled
            )
            async with self._drain_lock:
                # Use a lock to serialize drain() calls. Circumvents this bug:
                # https://bugs.python.org/issue29930
                await self.writer.drain()
        except (BrokenPipeError, ConnectionResetError):
            raise RemoteDisconnected()
        except RuntimeError as err:
            # We don't do a pre-check for a closed writer since this is a performance critical
            # path. We also don't want to swallow runtime errors unrelated to closed handlers.
            if "handler is closed" in str(err):
                self.logger.debug("Failed to send %s. Handler closed.", message)
                raise RemoteDisconnected from err
            raise

    async def read_message(self) -> Message:
        if self.reader.at_eof():
            raise RemoteDisconnected()

        try:
            raw_size = await self.reader.readexactly(constants.SIZE_MARKER_LENGTH)
            size = int.from_bytes(raw_size, "little")
            message = await self.reader.readexactly(size)
            obj = cast(Message, pickle.loads(message))
            return obj
        except (asyncio.IncompleteReadError, BrokenPipeError, ConnectionResetError):
            raise RemoteDisconnected()


class AsyncioRemoteEndpoint(BaseRemoteEndpoint):
    def __init__(
        self,
        local_name: str,
        conn: ConnectionAPI,
        subscriptions_changed: ConditionAPI,
        new_msg_func: Callable[[Broadcast], Awaitable[Any]],
    ) -> None:
        super().__init__(local_name, conn, new_msg_func)

        self._notify_lock = asyncio.Lock()  # type: ignore

        self._received_response = asyncio.Condition()  # type: ignore
        self._received_subscription = asyncio.Condition()  # type: ignore

        self._running = asyncio.Event()  # type: ignore
        self._stopped = asyncio.Event()  # type: ignore

        self._received_subscription = subscriptions_changed

        self._subscriptions_initialized = asyncio.Event()  # type: ignore

        self._running = asyncio.Event()  # type: ignore
        self._stopped = asyncio.Event()  # type: ignore
        self._ready = asyncio.Event()  # type: ignore

    #
    # Running API
    #
    @asynccontextmanager
    async def run(self) -> AsyncIterator[RemoteEndpointAPI]:
        await self._start()

        try:
            yield self
        finally:
            await self.stop()

    async def _start(self) -> None:
        self._task = asyncio.ensure_future(self._process_incoming_messages())
        await self.wait_started()

    async def stop(self) -> None:
        if self.is_stopped:
            return
        self._stopped.set()
        self._task.cancel()
        try:
            await self._task
        except asyncio.CancelledError:
            pass
        finally:
            await self.conn.close()


TFunc = TypeVar("TFunc", bound=Callable[..., Any])


SubscriptionAsyncHandler = Callable[[BaseEvent], Awaitable[Any]]
SubscriptionSyncHandler = Callable[[BaseEvent], Any]


class AsyncioEndpoint(BaseEndpoint):
    """
    The :class:`~lahja.endpoint.asyncio.AsyncioEndpoint` enables communication
    between different processes as well as within a single process via various
    event-driven APIs.
    """

    _ipc_path: Path

    _receiving_queue: "asyncio.Queue[Tuple[Union[bytes, BaseEvent], Optional[BroadcastConfig]]]"
    _receiving_loop_running: asyncio.Event

    _futures: Dict[RequestID, "asyncio.Future[BaseEvent]"]

    _async_handler: DefaultDict[Type[BaseEvent], List[SubscriptionAsyncHandler]]
    _sync_handler: DefaultDict[Type[BaseEvent], List[SubscriptionSyncHandler]]

    _loop: Optional[asyncio.AbstractEventLoop] = None

    _subscriptions_changed: asyncio.Event

    def __init__(self, name: str) -> None:
        super().__init__(name)

        # Signal when a new remote connection is established
        self._remote_connections_changed = asyncio.Condition()  # type: ignore

        # Signal when at least one remote has had a subscription change.
        self._remote_subscriptions_changed = asyncio.Condition()  # type: ignore

        # event for signaling local subscriptions have changed.
        self._subscriptions_changed = asyncio.Event()

        # storage for futures which are waiting for a response.
        self._futures: Dict[RequestID, "asyncio.Future[BaseEvent]"] = {}

        # handlers for event subscriptions.  These are
        # intentionally stored separately so that the cost of
        # `inspect.iscoroutine` is incurred once when the subscription is
        # created instead of for each event that is processed
        self._async_handler = defaultdict(list)
        self._sync_handler = defaultdict(list)

        # queues for stream handlers
        self._queues: Dict[Type[BaseEvent], List["asyncio.Queue[BaseEvent]"]] = {}

        # background tasks that are started as part of the process of running
        # the endpoint.
        self._endpoint_tasks: Set["asyncio.Future[Any]"] = set()

        # background tasks that are started as part of serving the endpoint
        # over an IPC socket.
        self._server_tasks: Set["asyncio.Future[Any]"] = set()

        self._running = False
        self._serving = False

    def __str__(self) -> str:
        return f"Endpoint[{self.name}]"

    def __repr__(self) -> str:
        return f"<{self.name}>"

    @property
    def is_running(self) -> bool:
        return self._running

    @property
    def is_serving(self) -> bool:
        return self._serving

    @property
    def ipc_path(self) -> Path:
        return self._ipc_path

    @property
    def event_loop(self) -> asyncio.AbstractEventLoop:
        if self._loop is None:
            raise AttributeError("Endpoint does not have an event loop set.")

        return self._loop

    def check_event_loop(func: TFunc) -> TFunc:  # type: ignore
        """
        All Endpoint methods must be called from the same event loop.
        """

        @functools.wraps(func)
        def run(self, *args, **kwargs):  # type: ignore
            if not self._loop:
                self._loop = asyncio.get_event_loop()

            if self._loop != asyncio.get_event_loop():
                raise Exception(
                    "All endpoint methods must be called from the same event loop"
                )

            return func(self, *args, **kwargs)

        return cast(TFunc, run)

    #
    # Running API
    #
    @asynccontextmanager
    async def run(self) -> AsyncIterator[EndpointAPI]:
        if not self._loop:
            self._loop = asyncio.get_event_loop()

        await self._start()

        try:
            yield self
        finally:
            await self._stop()

    @check_event_loop
    async def _start(self) -> None:
        if self.is_running:
            raise RuntimeError(f"Endpoint {self.name} is already running")
        self._receiving_loop_running = asyncio.Event()
        self._receiving_queue = asyncio.Queue()

        self._running = True

        self._endpoint_tasks.add(asyncio.ensure_future(self._connect_receiving_queue()))
        self._endpoint_tasks.add(
            asyncio.ensure_future(self._monitor_subscription_changes())
        )

        await self._receiving_loop_running.wait()
        self.logger.debug("Endpoint[%s]: running", self.name)

    async def _stop(self) -> None:
        """
        Stop the :class:`~lahja.endpoint.asyncio.AsyncioEndpoint` from receiving further events.
        """
        if not self.is_running:
            return

        await self._stop_server()

        self._running = False

        for task in tuple(self._endpoint_tasks):
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass

        self.logger.debug("Endpoint[%s]: stopped", self.name)

    async def _connect_receiving_queue(self) -> None:
        self._receiving_loop_running.set()
        while self.is_running:
            try:
                (item, config) = await self._receiving_queue.get()
            except RuntimeError as err:
                # do explicit check since RuntimeError is a bit generic and we
                # only want to catch the closed event loop case here.
                if str(err) == "Event loop is closed":
                    break
                raise
            try:
                event = self._decompress_event(item)
                await self._process_item(event, config)
            except Exception:
                traceback.print_exc()

    async def _monitor_subscription_changes(self) -> None:
        while self.is_running:
            # We wait for the event to change and then immediately replace it
            # with a new event.  This **must** occur before any additional
            # `await` calls to ensure that any *new* changes to the
            # subscriptions end up operating on the *new* event and will be
            # picked up in the next iteration of the loop.
            await self._subscriptions_changed.wait()
            self._subscriptions_changed = asyncio.Event()

            # make a copy so that the set doesn't change while we iterate
            # over it
            subscribed_events = self.get_subscribed_events()

            async with self._remote_connections_changed:
                await asyncio.gather(
                    *(
                        remote.notify_subscriptions_updated(
                            subscribed_events, block=False
                        )
                        for remote in self._connections
                    )
                )
            async with self._remote_subscriptions_changed:
                self._remote_subscriptions_changed.notify_all()

    async def _process_item(
        self, item: BaseEvent, config: Optional[BroadcastConfig]
    ) -> None:
        event_type = type(item)

        if config is not None and config.filter_event_id in self._futures:
            future = self._futures[config.filter_event_id]
            if not future.done():
                future.set_result(item)
            self._futures.pop(config.filter_event_id, None)

        if event_type in self._queues:
            for queue in self._queues[event_type]:
                queue.put_nowait(item)

        if event_type in self._sync_handler:
            for handler in self._sync_handler[event_type]:
                try:
                    handler(item)
                except Exception:
                    self.logger.debug(
                        "%s: Error in subscription handler %s",
                        self,
                        handler,
                        exc_info=True,
                    )

        if event_type in self._async_handler:
            await asyncio.gather(
                *(
                    self._run_async_subscription_handler(handler, item)
                    for handler in self._async_handler[event_type]
                )
            )

    async def _run_async_subscription_handler(
        self, handler: Callable[[BaseEvent], Awaitable[Any]], item: BaseEvent
    ) -> None:
        try:
            await handler(item)
        except Exception:
            self.logger.debug(
                "%s: Error in subscription handler %s", self, handler, exc_info=True
            )

    #
    # Server API
    #
    @classmethod
    @asynccontextmanager
    async def serve(cls, config: ConnectionConfig) -> AsyncIterator["AsyncioEndpoint"]:
        endpoint = cls(config.name)
        async with endpoint.run():
            await endpoint._start_server(config.path)
            try:
                yield endpoint
            finally:
                await endpoint._stop_server()

    @check_event_loop
    async def _start_server(self, ipc_path: Path) -> None:
        """
        Start serving this :class:`~lahja.endpoint.asyncio.AsyncioEndpoint` so that it
        can receive events. Await until the
        :class:`~lahja.endpoint.asyncio.AsyncioEndpoint` is ready.
        """
        if not self.is_running:
            raise RuntimeError(f"Endpoint {self.name} must be running to start server")
        elif self.is_serving:
            raise RuntimeError(f"Endpoint {self.name} is already serving")

        self._ipc_path = ipc_path

        self._serving = True

        self._server = await asyncio.start_unix_server(
            self._accept_conn, path=str(self.ipc_path)
        )
        self.logger.debug("Endpoint[%s]: server started", self.name)

    async def _stop_server(self) -> None:
        if not self.is_serving:
            return
        self._serving = False

        self._server.close()

        for task in tuple(self._server_tasks):
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass

        try:
            self.ipc_path.unlink()
        except FileNotFoundError:
            pass
        self.logger.debug("Endpoint[%s]: server stopped", self.name)

    async def _accept_conn(self, reader: StreamReader, writer: StreamWriter) -> None:
        conn = AsyncioConnection(reader, writer)
        remote = AsyncioRemoteEndpoint(
            local_name=self.name,
            conn=conn,
            subscriptions_changed=self._remote_subscriptions_changed,
            new_msg_func=self._receiving_queue.put,
        )

        self._server_tasks.add(asyncio.ensure_future(self._run_remote_endpoint(remote)))

    #
    # Connection API
    #
    def _throw_if_already_connected(self, *endpoints: ConnectionConfig) -> None:
        seen: Set[str] = set()

        for config in endpoints:
            if config.name == self.name:
                raise ConnectionAttemptRejected(
                    f"Cannot connect an endpoint to itself."
                )
            elif config.name in seen:
                raise ConnectionAttemptRejected(
                    f"Trying to connect to {config.name} twice. Names must be uniqe."
                )
            elif self.is_connected_to(config.name):
                raise ConnectionAttemptRejected(
                    f"Already connected to {config.name} at {config.path}. Names must be unique."
                )
            else:
                seen.add(config.name)

    @check_event_loop
    async def connect_to_endpoints(self, *endpoints: ConnectionConfig) -> None:
        """
        Connect to the given endpoints and await until all connections are established.
        """
        self._throw_if_already_connected(*endpoints)
        await asyncio.gather(
            *(self._await_connect_to_endpoint(endpoint) for endpoint in endpoints)
        )

    async def _await_connect_to_endpoint(self, endpoint: ConnectionConfig) -> None:
        self._throw_if_already_connected(endpoint)
        await wait_for_path(endpoint.path)

        conn = await AsyncioConnection.connect_to(endpoint.path)
        remote = AsyncioRemoteEndpoint(
            local_name=self.name,
            conn=conn,
            subscriptions_changed=self._remote_subscriptions_changed,
            new_msg_func=self._receiving_queue.put,
        )

        task = asyncio.ensure_future(self._run_remote_endpoint(remote))

        task.add_done_callback(self._endpoint_tasks.remove)

        self._endpoint_tasks.add(task)

        # block until we've received the other endpoint's subscriptions to
        # ensure that it is safe to broadcast
        await asyncio.wait_for(
            self.wait_until_connected_to(endpoint.name),
            timeout=constants.ENDPOINT_CONNECT_TIMEOUT,
        )

    #
    # Event API
    #
    async def broadcast(
        self, item: BaseEvent, config: Optional[BroadcastConfig] = None
    ) -> None:
        """
        Broadcast an instance of :class:`~lahja.common.BaseEvent` on the event bus. Takes
        an optional second parameter of :class:`~lahja.common.BroadcastConfig` to decide
        where this event should be broadcasted to. By default, events are broadcasted across
        all connected endpoints with their consuming call sites.
        """
        self.maybe_raise_no_subscribers_exception(config, type(item))
        await self._broadcast(item, config, None)

    async def _broadcast(
        self,
        item: BaseEvent,
        config: Optional[BroadcastConfig],
        id: Optional[RequestID],
    ) -> None:
        item.bind(self, id)

        is_eligible = should_endpoint_receive_item(
            item, config, self.name, self.get_subscribed_events()
        )
        is_internal = config is not None and config.internal

        if is_eligible:
            await self._process_item(item, config)

        if is_internal:
            # if the event is flagged as internal we exit early since it should
            # not be broadcast beyond this endpoint
            return

        # Broadcast to every connected Endpoint that is allowed to receive the event
        remotes_for_broadcast = tuple(
            remote
            for remote in self._connections
            if should_endpoint_receive_item(
                item, config, remote.name, remote.get_subscribed_events()
            )
        )

        compressed_item = self._compress_event(item)

        message = Broadcast(compressed_item, config)
        for remote in remotes_for_broadcast:
            try:
                await remote.send_message(message)
            except RemoteDisconnected:
                self.logger.debug(f"Remote endpoint {remote} no longer connected")
                await remote.stop()

    def broadcast_nowait(
        self, item: BaseEvent, config: Optional[BroadcastConfig] = None
    ) -> None:
        """
        A non-async :meth:`~lahja.Endpoint.broadcast` (see :meth:`~lahja.Endpoint.broadcast`
        for more)

        Instead of blocking the calling coroutine this function schedules the broadcast
        and immediately returns.

        CAUTION: You probably don't want to use this. broadcast() doesn't return until the
        write socket has finished draining, meaning that the OS has accepted the message.
        This prevents us from sending more data than the remote process can handle.
        broadcast_nowait has no such backpressure. Even after the remote process stops
        accepting new messages this function will continue to accept them, which in the
        worst case could lead to runaway memory usage.
        """
        self.maybe_raise_no_subscribers_exception(config, type(item))
        asyncio.ensure_future(self._broadcast(item, config, None))

    @check_event_loop
    async def request(
        self,
        item: BaseRequestResponseEvent[TResponse],
        config: Optional[BroadcastConfig] = None,
    ) -> TResponse:
        """
        Broadcast an instance of
        :class:`~lahja.common.BaseRequestResponseEvent` on the event bus and
        immediately wait on an expected answer of type
        :class:`~lahja.common.BaseEvent`. Optionally pass a second parameter of
        :class:`~lahja.common.BroadcastConfig` to decide where the request
        should be broadcasted to. By default, requests are broadcasted across
        all connected endpoints with their consuming call sites.
        """
        self.maybe_raise_no_subscribers_exception(config, type(item))
        request_id = next(self._get_request_id)

        future: "asyncio.Future[TResponse]" = asyncio.Future()
        self._futures[request_id] = future  # type: ignore
        # mypy can't reconcile the TResponse with the declared type of
        # `self._futures`.

        await self._broadcast(item, config, request_id)

        future.add_done_callback(
            functools.partial(self._remove_cancelled_future, request_id)
        )

        result = await future

        expected_response_type = item.expected_response_type()
        if not isinstance(result, expected_response_type):
            raise UnexpectedResponse(
                f"The type of the response is {type(result)}, expected: {expected_response_type}"
            )

        return result

    def _remove_cancelled_future(
        self, id: RequestID, future: "asyncio.Future[Any]"
    ) -> None:
        try:
            future.exception()
        except asyncio.CancelledError:
            self._futures.pop(id, None)

    def _remove_async_subscription(
        self, event_type: Type[BaseEvent], handler_fn: SubscriptionAsyncHandler
    ) -> None:
        self._async_handler[event_type].remove(handler_fn)
        if not self._async_handler[event_type]:
            self._async_handler.pop(event_type)
        # this is asynchronous because that's a better user experience than making
        # the user `await subscription.remove()`. This means this Endpoint will keep
        # getting events for a little while after it stops listening for them but
        # that's a performance problem, not a correctness problem.
        self._subscriptions_changed.set()

    def _remove_sync_subscription(
        self, event_type: Type[BaseEvent], handler_fn: SubscriptionSyncHandler
    ) -> None:
        self._sync_handler[event_type].remove(handler_fn)
        if not self._sync_handler[event_type]:
            self._sync_handler.pop(event_type)
        # this is asynchronous because that's a better user experience than making
        # the user `await subscription.remove()`. This means this Endpoint will keep
        # getting events for a little while after it stops listening for them but
        # that's a performance problem, not a correctness problem.
        self._subscriptions_changed.set()

    def subscribe(
        self,
        event_type: Type[TSubscribeEvent],
        handler: Callable[[TSubscribeEvent], Union[Any, Awaitable[Any]]],
    ) -> Subscription:
        """
        Subscribe to receive updates for any event that matches the specified event type.
        A handler is passed as a second argument an :class:`~lahja.common.Subscription` is returned
        to unsubscribe from the event if needed.
        """
        if asyncio.iscoroutinefunction(handler):
            casted_handler = cast(SubscriptionAsyncHandler, handler)
            self._async_handler[event_type].append(casted_handler)
            unsubscribe_fn = functools.partial(
                self._remove_async_subscription, event_type, casted_handler
            )
        else:
            casted_handler = cast(SubscriptionSyncHandler, handler)
            self._sync_handler[event_type].append(casted_handler)
            unsubscribe_fn = functools.partial(
                self._remove_sync_subscription, event_type, casted_handler
            )

        self._subscriptions_changed.set()

        return Subscription(unsubscribe_fn)

    async def stream(
        self, event_type: Type[TStreamEvent], num_events: Optional[int] = None
    ) -> AsyncGenerator[TStreamEvent, None]:
        """
        Stream all events that match the specified event type. This returns an
        ``AsyncIterable[BaseEvent]`` which can be consumed through an ``async for`` loop.
        An optional ``num_events`` parameter can be passed to stop streaming after a maximum amount
        of events was received.
        """
        queue: "asyncio.Queue[TStreamEvent]" = asyncio.Queue()
        casted_queue = cast("asyncio.Queue[BaseEvent]", queue)
        if event_type not in self._queues:
            self._queues[event_type] = []

        self._queues[event_type].append(casted_queue)
        self._subscriptions_changed.set()

        iterations: Iterable[int]

        if num_events is None:
            # loop forever
            iterations = itertools.count()
        else:
            iterations = range(num_events)

        try:
            for _ in iterations:
                try:
                    yield await queue.get()
                except GeneratorExit:
                    break
        finally:
            self._queues[event_type].remove(casted_queue)
            if not self._queues[event_type]:
                del self._queues[event_type]
            self._subscriptions_changed.set()

    #
    # Subscription API
    #
    def get_subscribed_events(self) -> Set[Type[BaseEvent]]:
        """
        Return the set of events this Endpoint is currently listening for
        """
        return (
            set(self._sync_handler.keys())
            .union(self._async_handler.keys())
            .union(self._queues.keys())
        )



from abc import ABC, abstractmethod
from typing import Any, Type

from lahja.asyncio import AsyncioEndpoint
from lahja.base import BaseEndpoint
from lahja.trio import TrioEndpoint


class BaseBackend(ABC):
    name: str
    Endpoint: Type[BaseEndpoint]

    @staticmethod
    @abstractmethod
    def run(coro: Any, *args: Any) -> None:
        pass

    @staticmethod
    @abstractmethod
    async def sleep(seconds: float) -> None:
        pass


class AsyncioBackend(BaseBackend):
    name = "asyncio"
    Endpoint = AsyncioEndpoint

    @staticmethod
    def run(coro: Any, *args: Any) -> None:
        # UNCOMMENT FOR DEBUGGING
        # logger = multiprocessing.log_to_stderr()
        # logger.setLevel(logging.INFO)
        import asyncio

        loop = asyncio.get_event_loop()
        loop.run_until_complete(coro(*args))
        loop.stop()

    @staticmethod
    async def sleep(seconds: float) -> None:
        import asyncio

        await asyncio.sleep(seconds)


class TrioBackend(BaseBackend):
    name = "trio"
    Endpoint = TrioEndpoint

    @staticmethod
    def run(coro: Any, *args: Any) -> None:
        # UNCOMMENT FOR DEBUGGING
        # logger = multiprocessing.log_to_stderr()
        # logger.setLevel(logging.INFO)
        import trio

        trio.run(coro, *args)

    @staticmethod
    async def sleep(seconds: float) -> None:
        import trio

        await trio.sleep(seconds)



from abc import ABC, abstractmethod
import logging
from pathlib import Path
import pickle
from typing import (  # noqa: F401
    Any,
    AsyncContextManager,
    AsyncGenerator,
    Awaitable,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from ._snappy import check_has_snappy_support
from .common import (
    BaseEvent,
    BaseRequestResponseEvent,
    Broadcast,
    BroadcastConfig,
    ConnectionConfig,
    Hello,
    Message,
    Msg,
    RequestIDGenerator,
    Subscription,
    SubscriptionsAck,
    SubscriptionsUpdated,
)
from .exceptions import ConnectionAttemptRejected, NoSubscribers, RemoteDisconnected
from .typing import ConditionAPI, EventAPI, LockAPI, RequestID

TResponse = TypeVar("TResponse", bound=BaseEvent)
TWaitForEvent = TypeVar("TWaitForEvent", bound=BaseEvent)
TSubscribeEvent = TypeVar("TSubscribeEvent", bound=BaseEvent)
TStreamEvent = TypeVar("TStreamEvent", bound=BaseEvent)


class ConnectionAPI(ABC):
    @classmethod
    @abstractmethod
    async def connect_to(cls, path: Path) -> "ConnectionAPI":
        ...

    @abstractmethod
    async def close(self) -> None:
        ...

    @abstractmethod
    async def send_message(self, message: Msg) -> None:
        ...

    @abstractmethod
    async def read_message(self) -> Message:
        ...


class RemoteEndpointAPI(ABC):
    """
    Represents a connection to another endpoint.  Connections *can* be
    bi-directional with messages flowing in either direction.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        ...

    #
    # Object run lifecycle
    #
    @abstractmethod
    async def wait_started(self) -> None:
        ...

    @abstractmethod
    async def wait_ready(self) -> None:
        ...

    @abstractmethod
    async def wait_stopped(self) -> None:
        ...

    @property
    @abstractmethod
    def is_running(self) -> bool:
        ...

    @property
    @abstractmethod
    def is_ready(self) -> bool:
        ...

    @property
    @abstractmethod
    def is_stopped(self) -> bool:
        ...

    #
    # Running API
    #
    @abstractmethod
    def run(self) -> AsyncContextManager["RemoteEndpointAPI"]:
        """
        Context manager API for running endpoints.

        .. code-block:: python

            async with endpoint.run() as endpoint:
                ... # endpoint running within context
            ... # endpoint stopped after

        test
        """
        ...

    @abstractmethod
    async def stop(self) -> None:
        """
        Manually stop the remote endpoint.
        """
        ...

    #
    # Core external API
    #
    @abstractmethod
    def get_subscribed_events(self) -> Set[Type[BaseEvent]]:
        """
        Return the event types this endpoint is current subscribed to.
        """
        ...

    @abstractmethod
    async def notify_subscriptions_updated(
        self, subscriptions: Set[Type[BaseEvent]], block: bool = True
    ) -> None:
        """
        Alert the endpoint on the other side of this connection that the local
        subscriptions have changed. If ``block`` is ``True`` then this function
        will block until the remote endpoint has acknowledged the new
        subscription set. If ``block`` is ``False`` then this function will
        return immediately after the send finishes.
        """
        ...

    @abstractmethod
    async def send_message(self, message: Msg) -> None:
        ...

    @abstractmethod
    async def wait_until_subscription_initialized(self) -> None:
        """
        Block until at least one `SubscriptionsUpdated` event has been
        received.
        """
        ...


class BaseRemoteEndpoint(RemoteEndpointAPI):
    """
    Base class to implement common logic that can be shared across different
    implementations of the `RemoteEndpointAPI`

    Represents a connection to another endpoint.  Connections *can* be
    bi-directional with messages flowing in either direction.
    """

    logger = logging.getLogger("lahja.base.RemoteEndpoint")

    conn: ConnectionAPI

    _local_name: str
    _running: EventAPI
    _ready: EventAPI
    _stopped: EventAPI

    _notify_lock: LockAPI

    _received_response: ConditionAPI
    _received_subscription: ConditionAPI

    _subscribed_events: Set[Type[BaseEvent]]

    _subscriptions_initialized: EventAPI

    def __init__(
        self,
        local_name: str,
        conn: ConnectionAPI,
        new_msg_func: Callable[[Broadcast], Awaitable[Any]],
    ) -> None:
        self._local_name = local_name
        self.conn = conn
        self.new_msg_func = new_msg_func

        self._subscribed_events = set()

    def __str__(self) -> str:
        return f"RemoteEndpoint[{self._name if self._name is not None else id(self)}]"

    def __repr__(self) -> str:
        return f"<{self}>"

    _name: Optional[str] = None

    @property
    def name(self) -> str:
        if self._name is None:
            raise AttributeError("Endpoint name has not yet been established")
        else:
            return self._name

    async def wait_started(self) -> None:
        await self._running.wait()

    async def wait_ready(self) -> None:
        await self._ready.wait()

    async def wait_stopped(self) -> None:
        await self._stopped.wait()

    @property
    def is_running(self) -> bool:
        return not self.is_stopped and self._running.is_set()

    @property
    def is_ready(self) -> bool:
        return self.is_running and self._ready.is_set()

    @property
    def is_stopped(self) -> bool:
        return self._stopped.is_set()

    async def _process_incoming_messages(self) -> None:
        self._running.set()

        try:
            # Send the hello message
            await self.send_message(Hello(self._local_name))
            # Wait for the other endpoint to identify itself.
            hello = await self.conn.read_message()
        except RemoteDisconnected:
            self._stopped.set()
            return

        if isinstance(hello, Hello):
            self._name = hello.name
            self.logger.debug(
                "RemoteEndpoint connection established: %s <-> %s",
                self._local_name,
                self.name,
            )
            self._ready.set()
        else:
            self.logger.debug(
                "Invalid message: First message must be `Hello`, Got: %s", hello
            )
            self._stopped.set()
            return

        while self.is_running:
            try:
                message = await self.conn.read_message()
            except RemoteDisconnected:
                async with self._received_response:
                    self._received_response.notify_all()
                return

            if isinstance(message, Broadcast):
                await self.new_msg_func(message)
            elif isinstance(message, SubscriptionsUpdated):
                self._subscriptions_initialized.set()
                async with self._received_subscription:
                    self._subscribed_events = message.subscriptions
                    self._received_subscription.notify_all()
                # The ack is sent after releasing the lock since we've already
                # exited the code which actually updates the subscriptions and
                # we are merely responding to the sender to acknowledge
                # receipt.
                if message.response_expected:
                    await self.send_message(SubscriptionsAck())
            elif isinstance(message, SubscriptionsAck):
                async with self._received_response:
                    self._received_response.notify_all()
            else:
                self.logger.error(f"received unexpected message: {message}")

    def get_subscribed_events(self) -> Set[Type[BaseEvent]]:
        return self._subscribed_events

    async def notify_subscriptions_updated(
        self, subscriptions: Set[Type[BaseEvent]], block: bool = True
    ) -> None:
        """
        Alert the endpoint on the other side of this connection that the local
        subscriptions have changed. If ``block`` is ``True`` then this function
        will block until the remote endpoint has acknowledged the new
        subscription set. If ``block`` is ``False`` then this function will
        return immediately after the send finishes.
        """
        # The extra lock ensures only one coroutine can notify this endpoint at any one time
        # and that no replies are accidentally received by the wrong
        # coroutines. Without this, in the case where `block=True`, this inner
        # block would release the lock on the call to `wait()` which would
        # allow the ack from a different update to incorrectly result in this
        # returning before the ack had been received.
        async with self._notify_lock:
            async with self._received_response:
                try:
                    await self.conn.send_message(
                        SubscriptionsUpdated(subscriptions, block)
                    )
                except RemoteDisconnected:
                    return
                if block:
                    await self._received_response.wait()

    async def send_message(self, message: Msg) -> None:
        await self.conn.send_message(message)

    async def wait_until_subscription_initialized(self) -> None:
        await self._subscriptions_initialized.wait()


class EndpointAPI(ABC):
    """
    The :class:`~lahja.endpoint.Endpoint` enables communication between different processes
    as well as within a single process via various event-driven APIs.
    """

    __slots__ = ("name",)

    name: str

    @property
    @abstractmethod
    def is_running(self) -> bool:
        ...

    @property
    @abstractmethod
    def is_serving(self) -> bool:
        ...

    #
    # Running API
    #
    @abstractmethod
    def run(self) -> AsyncContextManager["EndpointAPI"]:
        """
        Context manager API for running endpoints.

        .. code-block:: python

            async with endpoint.run() as endpoint:
                ... # endpoint running within context
            ... # endpoint stopped after

        """
        ...

    #
    # Serving API
    #
    @classmethod
    @abstractmethod
    def serve(cls, config: ConnectionConfig) -> AsyncContextManager["EndpointAPI"]:
        """
        Context manager API for running and endpoint server.

        .. code-block:: python

            async with EndpointClass.serve(config):
                ... # server running within context
            ... # server stopped

        """
        ...

    #
    # Connection API
    #

    @abstractmethod
    async def connect_to_endpoints(self, *endpoints: ConnectionConfig) -> None:
        """
        Establish connections to the given endpoints.
        """
        ...

    @abstractmethod
    def is_connected_to(self, endpoint_name: str) -> bool:
        """
        Return whether this endpoint is connected to another endpoint with the given name.
        """
        ...

    @abstractmethod
    async def wait_until_connected_to(self, endpoint_name: str) -> None:
        """
        Return once a connection exists to an endpoint with the given name.
        """
        ...

    @abstractmethod
    def get_connected_endpoints_and_subscriptions(
        self
    ) -> Tuple[Tuple[str, Set[Type[BaseEvent]]], ...]:
        """
        Return 2-tuples for all all connected endpoints containing the name of
        the endpoint coupled with the set of messages the endpoint subscribes
        to
        """
        ...

    @abstractmethod
    async def wait_until_connections_change(self) -> None:
        """
        Block until the set of connected remote endpoints changes.
        """
        ...

    #
    # Event API
    #
    @abstractmethod
    async def broadcast(
        self, item: BaseEvent, config: Optional[BroadcastConfig] = None
    ) -> None:
        """
        Broadcast an instance of :class:`~lahja.common.BaseEvent` on the event bus. Takes
        an optional second parameter of :class:`~lahja.common.BroadcastConfig` to decide
        where this event should be broadcasted to. By default, events are broadcasted across
        all connected endpoints with their consuming call sites.
        """
        ...

    @abstractmethod
    def broadcast_nowait(
        self, item: BaseEvent, config: Optional[BroadcastConfig] = None
    ) -> None:
        """
        A sync compatible version of :meth:`~lahja.base.EndpointAPI.broadcast`

        .. warning::

            Heavy use of :meth:`~lahja.base.EndpointAPI.broadcast_nowait` in
            contiguous blocks of code without yielding to the `async`
            implementation should be expected to cause problems.

        """
        ...

    @abstractmethod
    async def request(
        self,
        item: BaseRequestResponseEvent[TResponse],
        config: Optional[BroadcastConfig] = None,
    ) -> TResponse:
        """
        Broadcast an instance of
        :class:`~lahja.common.BaseRequestResponseEvent` on the event bus and
        immediately wait on an expected answer of type
        :class:`~lahja.common.BaseEvent`. Optionally pass a second parameter of
        :class:`~lahja.common.BroadcastConfig` to decide where the request
        should be broadcasted to. By default, requests are broadcasted across
        all connected endpoints with their consuming call sites.
        """
        ...

    @abstractmethod
    def subscribe(
        self,
        event_type: Type[TSubscribeEvent],
        handler: Callable[[TSubscribeEvent], Union[Any, Awaitable[Any]]],
    ) -> Subscription:
        """
        Subscribe to receive updates for any event that matches the specified event type.
        A handler is passed as a second argument an :class:`~lahja.common.Subscription` is returned
        to unsubscribe from the event if needed.
        """
        ...

    @abstractmethod
    async def stream(
        self, event_type: Type[TStreamEvent], num_events: Optional[int] = None
    ) -> AsyncGenerator[TStreamEvent, None]:
        """
        Stream all events that match the specified event type. This returns an
        ``AsyncIterable[BaseEvent]`` which can be consumed through an ``async for`` loop.
        An optional ``num_events`` parameter can be passed to stop streaming after a maximum amount
        of events was received.
        """
        yield  # type: ignore  # yield statemen convinces mypy this is a generator function

    @abstractmethod
    async def wait_for(self, event_type: Type[TWaitForEvent]) -> TWaitForEvent:
        """
        Wait for a single instance of an event that matches the specified event type.
        """
        ...

    #
    # Subscription API
    #
    @abstractmethod
    def get_subscribed_events(self) -> Set[Type[BaseEvent]]:
        """
        Return the set of event types this endpoint subscribes to.
        """
        ...

    @abstractmethod
    async def wait_until_endpoint_subscriptions_change(self) -> None:
        """
        Block until any subscription change occurs on any remote endpoint or
        the set of remote endpoints changes
        """
        ...

    @abstractmethod
    def is_endpoint_subscribed_to(
        self, remote_endpoint: str, event_type: Type[BaseEvent]
    ) -> bool:
        """
        Return ``True`` if the specified remote endpoint is subscribed to the specified event type
        from this endpoint. Otherwise return ``False``.
        """
        ...

    @abstractmethod
    def is_any_endpoint_subscribed_to(self, event_type: Type[BaseEvent]) -> bool:
        """
        Return ``True`` if at least one of the connected remote endpoints is subscribed to the
        specified event type from this endpoint. Otherwise return ``False``.
        """
        ...

    @abstractmethod
    def are_all_endpoints_subscribed_to(self, event_type: Type[BaseEvent]) -> bool:
        """
        Return ``True`` if every connected remote endpoint is subscribed to the specified event
        type from this endpoint. Otherwise return ``False``.
        """
        ...

    @abstractmethod
    async def wait_until_endpoint_subscribed_to(
        self, remote_endpoint: str, event: Type[BaseEvent]
    ) -> None:
        """
        Block until the specified remote endpoint has subscribed to the specified event type
        from this endpoint.
        """
        ...

    @abstractmethod
    async def wait_until_any_endpoint_subscribed_to(
        self, event: Type[BaseEvent]
    ) -> None:
        """
        Block until any other remote endpoint has subscribed to the specified event type
        from this endpoint.
        """
        ...

    @abstractmethod
    async def wait_until_all_endpoints_subscribed_to(
        self, event: Type[BaseEvent], *, include_self: bool = True
    ) -> None:
        """
        Block until all currently connected remote endpoints are subscribed to the specified
        event type from this endpoint.
        """
        ...


class BaseEndpoint(EndpointAPI):
    """
    Base class for endpoint implementations that implements shared/common logic
    """

    _remote_connections_changed: ConditionAPI
    _remote_subscriptions_changed: ConditionAPI

    _connections: Set[RemoteEndpointAPI]

    _get_request_id: Iterator[RequestID]

    logger = logging.getLogger("lahja.endpoint.Endpoint")

    def __init__(self, name: str) -> None:
        self.name = name

        try:
            self._get_request_id = RequestIDGenerator(name.encode("ascii") + b":")
        except UnicodeDecodeError:
            raise Exception(
                f"TODO: Invalid endpoint name: '{name}'. Must be ASCII encodable string"
            )

        # storage containers for inbound and outbound connections to other
        # endpoints
        self._connections = set()

    def __str__(self) -> str:
        return f"Endpoint[{self.name}]"

    def __repr__(self) -> str:
        return f"<{self.name}>"

    #
    # Common implementations
    #
    def __reduce__(self) -> None:  # type: ignore
        raise NotImplementedError("Endpoints cannot be pickled")

    #
    # Connection API
    #
    def is_connected_to(self, endpoint_name: str) -> bool:
        if endpoint_name == self.name:
            return True
        return any(endpoint_name == remote.name for remote in self._connections)

    async def wait_until_connected_to(self, endpoint_name: str) -> None:
        """
        Return once a connection exists to an endpoint with the given name.
        """
        async with self._remote_connections_changed:
            while True:
                if self.is_connected_to(endpoint_name):
                    return
                await self._remote_connections_changed.wait()

    def get_connected_endpoints_and_subscriptions(
        self
    ) -> Tuple[Tuple[str, Set[Type[BaseEvent]]], ...]:
        """
        Return all connected endpoints and their event type subscriptions to this endpoint.
        """
        return ((self.name, self.get_subscribed_events()),) + tuple(
            (remote.name, remote.get_subscribed_events())
            for remote in self._connections
        )

    def maybe_raise_no_subscribers_exception(
        self, config: Optional[BroadcastConfig], event_type: Type[BaseEvent]
    ) -> None:
        """
        Check the given ``config`` and ``event_type`` and raise a
        :class:`~lahja.exceptions.NoSubscriber` if no subscribers exist for the ``event_type``
        when they at least one subscriber is expected.
        """

        if config is not None:
            if config.require_subscriber:
                return
            if config.filter_event_id is not None:
                # This is a response to a request
                return
        elif not self.is_any_endpoint_subscribed_to(event_type):
            raise NoSubscribers(f"No subscribers for: {event_type}")

    async def wait_until_connections_change(self) -> None:
        """
        Block until the set of connected remote endpoints changes.
        """
        async with self._remote_connections_changed:
            await self._remote_connections_changed.wait()

    #
    # Event API
    #
    async def wait_for(self, event_type: Type[TWaitForEvent]) -> TWaitForEvent:
        """
        Wait for a single instance of an event that matches the specified event type.
        """
        agen = self.stream(event_type, num_events=1)
        event = await agen.asend(None)
        await agen.aclose()
        return event

    #
    # Subscription API
    #
    async def wait_until_endpoint_subscriptions_change(self) -> None:
        async with self._remote_subscriptions_changed:
            await self._remote_subscriptions_changed.wait()

    def is_endpoint_subscribed_to(
        self, remote_endpoint: str, event_type: Type[BaseEvent]
    ) -> bool:
        """
        Return ``True`` if the specified remote endpoint is subscribed to the specified event type
        from this endpoint. Otherwise return ``False``.
        """
        for endpoint, subscriptions in self.get_connected_endpoints_and_subscriptions():
            if endpoint != remote_endpoint:
                continue

            for subscribed_event_type in subscriptions:
                if subscribed_event_type is event_type:
                    return True

        return False

    def is_any_endpoint_subscribed_to(self, event_type: Type[BaseEvent]) -> bool:
        """
        Return ``True`` if at least one of the connected remote endpoints is subscribed to the
        specified event type from this endpoint. Otherwise return ``False``.
        """
        for endpoint, subscriptions in self.get_connected_endpoints_and_subscriptions():
            for subscribed_event_type in subscriptions:
                if subscribed_event_type is event_type:
                    return True

        return False

    def are_all_endpoints_subscribed_to(
        self, event_type: Type[BaseEvent], include_self: bool = True
    ) -> bool:
        """
        Return ``True`` if every connected remote endpoint is subscribed to the specified event
        type from this endpoint. Otherwise return ``False``.
        """
        for endpoint, subscriptions in self.get_connected_endpoints_and_subscriptions():
            if not include_self and endpoint == self.name:
                continue

            if event_type not in subscriptions:
                return False

        return True

    async def wait_until_endpoint_subscribed_to(
        self, remote_endpoint: str, event: Type[BaseEvent]
    ) -> None:
        """
        Block until the specified remote endpoint has subscribed to the specified event type
        from this endpoint.
        """
        async with self._remote_subscriptions_changed:
            while True:
                if self.is_endpoint_subscribed_to(remote_endpoint, event):
                    return
                await self._remote_subscriptions_changed.wait()

    async def wait_until_any_endpoint_subscribed_to(
        self, event: Type[BaseEvent]
    ) -> None:
        """
        Block until any other remote endpoint has subscribed to the specified event type
        from this endpoint.
        """
        async with self._remote_subscriptions_changed:
            while True:
                if self.is_any_endpoint_subscribed_to(event):
                    return
                await self._remote_subscriptions_changed.wait()

    async def wait_until_all_endpoints_subscribed_to(
        self, event: Type[BaseEvent], *, include_self: bool = True
    ) -> None:
        """
        Block until all currently connected remote endpoints are subscribed to the specified
        event type from this endpoint.
        """
        async with self._remote_subscriptions_changed:
            while True:
                if self.are_all_endpoints_subscribed_to(
                    event, include_self=include_self
                ):
                    return
                await self._remote_subscriptions_changed.wait()

    #
    # Compression
    #

    # This property gets assigned during class creation.  This should be ok
    # since snappy support is defined as the module being importable and that
    # should not change during the lifecycle of the python process.
    has_snappy_support = check_has_snappy_support()

    def _compress_event(self, event: BaseEvent) -> Union[BaseEvent, bytes]:
        if self.has_snappy_support:
            import snappy

            return cast(
                bytes,
                snappy.compress(pickle.dumps(event, protocol=pickle.HIGHEST_PROTOCOL)),
            )
        else:
            return event

    def _decompress_event(self, data: Union[BaseEvent, bytes]) -> BaseEvent:
        if isinstance(data, BaseEvent):
            return data
        else:
            import snappy

            return cast(BaseEvent, pickle.loads(snappy.decompress(data)))

    #
    # Remote Endpoint Management
    #
    async def _run_remote_endpoint(self, remote: RemoteEndpointAPI) -> None:
        async with remote.run():
            await remote.wait_ready()
            await self._add_connection(remote)
            await remote.wait_until_subscription_initialized()
            await remote.wait_stopped()

        if remote in self._connections:
            await self._remove_connection(remote)

    async def _add_connection(self, remote: RemoteEndpointAPI) -> None:
        if remote in self._connections:
            raise ConnectionAttemptRejected("Remote is already tracked")
        elif self.is_connected_to(remote.name):
            raise ConnectionAttemptRejected(
                f"Already connected to remote with name {remote.name}"
            )

        async with self._remote_connections_changed:
            # then add them to our set of connections and signal that both
            # the remote connections and subscriptions have been updated.
            await remote.notify_subscriptions_updated(self.get_subscribed_events())
            self._connections.add(remote)
            self._remote_connections_changed.notify_all()
        async with self._remote_subscriptions_changed:
            self._remote_subscriptions_changed.notify_all()

    async def _remove_connection(self, remote: RemoteEndpointAPI) -> None:
        async with self._remote_connections_changed:
            self._connections.remove(remote)
            self._remote_connections_changed.notify_all()
        async with self._remote_subscriptions_changed:
            self._remote_subscriptions_changed.notify_all()



from abc import ABC, abstractmethod
from types import TracebackType
from typing import Any, Callable, NewType, Type

RequestID = NewType("RequestID", bytes)


class LockAPI(ABC):
    @abstractmethod
    async def __aenter__(self) -> None:
        ...

    @abstractmethod
    async def __aexit__(
        self, exc_type: Type[BaseException], exc_val: BaseException, tb: TracebackType
    ) -> None:
        ...

    @abstractmethod
    async def acquire(self) -> None:
        ...

    @abstractmethod
    def locked(self) -> bool:
        ...

    @abstractmethod
    def release(self) -> None:
        ...


class ConditionAPI(ABC):
    @abstractmethod
    async def __aenter__(self) -> None:
        ...

    @abstractmethod
    async def __aexit__(
        self, exc_type: Type[BaseException], exc_val: BaseException, tb: TracebackType
    ) -> None:
        ...

    @abstractmethod
    async def acquire(self) -> None:
        ...

    @abstractmethod
    def notify_all(self) -> None:
        ...

    @abstractmethod
    def notify(self, n: int) -> None:
        ...

    @abstractmethod
    def release(self, n: int) -> None:
        ...

    @abstractmethod
    async def wait(self) -> None:
        ...

    @abstractmethod
    async def wait_for(self, predicate: Callable[[], Any]) -> None:
        ...


class EventAPI(ABC):
    @abstractmethod
    def is_set(self) -> bool:
        ...

    @abstractmethod
    def set(self) -> None:
        ...

    @abstractmethod
    def clear(self) -> None:
        ...

    @abstractmethod
    async def wait(self) -> None:
        ...



from typing import Tuple

from lahja import ConnectionConfig


def create_consumer_endpoint_configs(
    num_processes: int
) -> Tuple[ConnectionConfig, ...]:
    return tuple(
        ConnectionConfig.from_name(create_consumer_endpoint_name(i))
        for i in range(num_processes)
    )


def create_consumer_endpoint_name(id: int) -> str:
    return f"consumer_{id}"
